#include <stdlib.h>

double c_canberra(double *x, double *y, long n);
double c_canberra_location(long *x, long *y, long n, long k);
double c_canberra_stability(long *x, long n, long p, long k);
double c_canberra_expected(long n, long k);
